@@ -77,13 +77,11 @@ class NeoStat(SimpleNamespace):
 
         # Renamed attrs
         self.name = getattr(self._data_, 'ZONE_NAME', getattr(self._data_, 'device', None))
-        self.floor_limit = getattr(self._data_, 'ZONE_NAME', None)
         self.target_temperature = getattr(self._data_, 'SET_TEMP', None)
         self.temperature = getattr(self._data_, 'ACTUAL_TEMP', None)
 
         # must be ints
         self.pin_number = int(self.pin_number)
-        self.preheat_active = int(self.preheat_active)
 
         # HOLD_TIME can be up to 99:99
         _hold_time = list(map(int, self.hold_time.split(':')))
